/* program cpxnet: read a file in pprn format, creates an equivalent linear program,
   and solves it through CPXhybnetopt (CPXnetopt + CPXdualopt)
   The problem is read from stardand input (cpxnet works as a filter).
   The user chooses the solution method by argv[1]. The possibilities are:'d'= network 
   problem+dual simplex; 'p'= network problem+primal simplex; 'b'= barrier optimizer 
   (Interior Point method). If desired, a name for identifying the problem can be passed 
   by argv[2].
   Mutual capacity constraints with an inf=1.0e+32 value are considered inactive and
   removed
   */


#include "cplex.h"

/* Bring in the declarations for the string and character functions 
   and malloc */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/times.h>

/* system dependent macros */
#ifdef LINUX
#define CLOCKS_PER_SECOND CLOCKS_PER_SEC  /* for Linux */
#endif
#ifdef ALPHA
#define CLOCKS_PER_SECOND 60.0  /* for OSF-alpha */
#endif
#ifdef SUN
#define CLOCKS_PER_SECOND 100.0  /* for Solaris */
#endif

/* macros */
#define ALLOC(ptr,ne,type) if (!((ptr)= (type *)calloc((ne),sizeof(type)))) \
                           {fprintf(stderr, "calloc failed\n"); exit(-1);} 

#define REALLOC(ptr,n,type) if (!((ptr)= (type *)realloc(((char *)ptr),(n)*sizeof(type)))) \
                            {fprintf(stderr, "realloc failed\n"); exit(-1);} 

#define SWAP(x,y,tmp) (tmp)=(x);(x)=(y);(y)=(tmp);

/* function prototypes */
static void read_net_create_lp(int *numcols_p, int * numrows_p, int * objsen_p, 
			       double **obj_p, double **rhs_p, char **sense_p, 
			       int **matbeg_p, int **matcnt_p, int **matind_p, 
			       double **matval_p, double **lb_p, double **ub_p, 
			       double **rngval_p, int *colspace_p, int *rowspace_p, 
			       int *nzspace_p);
static void freereaddata  (char **probname_p, double **obj_p, double **rhs_p, 
               char **sense_p, int **matbeg_p, int **matcnt_p, 
               int **matind_p, double **matval_p, double **lb_p, 
               double **ub_p, double **rngval_p,
               char ***colname_p, char **colnamestore_p, 
               char ***rowname_p, char **rownamestore_p);
static void free_and_null (char **ptr);
static void free_and_null2 (char *ptr);
static void write_double(double *v, int i, int j, char *nom);
static void write_int(int *v, int i, int j, char *nom);
static void write_char(char *v, int i, int j, char *nom);
void read_dimen_pprn_(int *, int *, int *, int *, int *); 
void read_format_pprn_(int *, int *, int *, int *, int *, double *, double *,
		      double *, double *, double *, double *, double *, 
		      int *, int *, int *, int *, int *);

int main (int argc, char *argv[])
{
  /* Declare and allocate space for the variables and arrays that will contain
   the data which define the LP problem */
  
  char     *probname = NULL;
  int      numcols   = 0;
  int      numrows   = 0;
  int      objsen    = 0;
  double   *obj      = NULL;
  double   *rhs      = NULL;
  char     *sense    = NULL;
  int      *matbeg   = NULL;
  int      *matcnt   = NULL;
  int      *matind   = NULL;
  double   *matval   = NULL;
  double   *lb       = NULL;
  double   *ub       = NULL;
  double   *rngval   = NULL;
  char     **colname = NULL;
  char     *colnamestore = NULL;
  char     **rowname = NULL;
  char     *rownamestore = NULL;
  int      colspace  = 0;
  int      rowspace  = 0;
  int      nzspace   = 0;
  unsigned colnamespace = 0;
  unsigned rownamespace = 0;

  /* Declare and allocate space for the variables and arrays where we will
     store the optimization results including the status, objective value,
     variable values, dual values, row slacks and variable reduced costs. */

  int      solstat;
  double   objval;
  double   *x     = NULL;
  int      *cstat = NULL;
  int      *rstat = NULL;
  int      nit;


  CPXENVptr     env = NULL;
  CPXLPptr      lp = NULL;
  int           status;
  int           j;
  int           cur_numrows, cur_numcols;

  char          *basismsg;

  /* declare variables for printing results of the optimization */
  double cputime;
  struct tms buffer;
  FILE *fout;



  /* Initialize the CPLEX environment */

  env = CPXopenCPLEX (&status);

  /* If an error occurs, the status value indicates the reason for
     failure.  A call to CPXgeterrorstring will produce the text of
     the error message.  Note that CPXopenCPLEX produces no output,
     so the only way to see the cause of the error is to use
     CPXgeterrorstring.  For other CPLEX routines, the errors will
     be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */

  if ( env == NULL ) {
    char  errmsg[1024];
    fprintf (stderr, "Could not open CPLEX environment.\n");
    CPXgeterrorstring (env, status, errmsg);
    fprintf (stderr, "%s", errmsg);
    goto TERMINATE;
  }
  
  /* Turn on output to the screen */

  status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
  if ( status != 0 ) {
    fprintf (stderr, 
	     "Failure to turn on screen indicator, error %d.\n", status);
    goto TERMINATE;
  }

  /* Increase Feasibility Tolerance */

/*
  status = CPXsetdblparam (env, CPX_PARAM_EPRHS, 0.1e-5);
  if ( status != 0 ) {
    fprintf (stderr, 
	     "Failure to increase feasibility tolerance, error %d.\n", status);
    goto TERMINATE;
  }
*/

  /* read network problem in pprn format and return linear problem dimensions and
     vectors */
  
  read_net_create_lp(&numcols, &numrows, &objsen, &obj, &rhs, &sense, &matbeg, &matcnt, 
		     &matind, &matval, &lb, &ub, &rngval, &colspace, &rowspace,
		     &nzspace);  

  /* write problem structure */
  /*  
  printf("numcols= %d  numrows= %d  colspace= %d  rowspace= %d  nzspace= %d\n",
	 numcols,numrows,colspace,rowspace,nzspace);
  write_double(obj,0,numcols-1,"obj");
  write_double(lb,0,numcols-1,"lb");
  write_double(ub,0,numcols-1,"ub");
  write_double(rhs,0,numrows-1,"rhs");
  write_char(sense,0,numrows-1,"sense");
  if (rngval != NULL) write_double(rngval,0,numrows-1,"rngval");
  write_int(matbeg,0,numcols-1,"matbeg");
  write_int(matcnt,0,numcols-1,"matcnt");
  write_int(matind,0,nzspace-1,"matind");
  write_double(matval,0,nzspace-1,"matval");
  */

  /* Load the problem. */
  lp = CPXloadlp (env, argv[2], numcols, numrows, objsen, 
		  obj, rhs, sense, matbeg, matcnt, matind, 
		  matval, lb, ub, rngval, colspace, rowspace,
		  nzspace);
  

  /* A returned pointer of NULL may mean that not enough memory
     was available or there was some other problem.  In the case of 
     failure, an error message will have been written to the error 
     channel from inside CPLEX.  In this example, the setting of
     the parameter CPX_PARAM_SCRIND causes the error message to
     appear on stdout.  Note that most CPLEX routines return
     an error code to indicate the reason for failure.  An error
     message is always printed to the error channel in these cases
     as well.  */
  
  if ( lp == NULL ) {
    fprintf (stderr, "Failed to load LP.\n");
    goto TERMINATE;
  }
  
  /* deactivation of some indicators */

  /* deactivate the presolve indicator */
  /*
  status = CPXsetintparam (env, CPX_PARAM_PREIND, CPX_OFF);
  if ( status != 0 ) {
    fprintf (stderr,"Failure to deactivate presolve indicator, error %d.\n", status);
    goto TERMINATE;
  }
  */

  /* deactivate the advanced start indicator */
  /*
  status = CPXsetintparam (env, CPX_PARAM_ADVIND, CPX_OFF);
  if ( status != 0 ) {
    fprintf (stderr,"Failure to deactivate advanced start indicator, error %d.\n", status);
    goto TERMINATE;
  }
  */

  /* deactivate the agregator indicator */
  /*
  status = CPXsetintparam (env, CPX_PARAM_AGGIND, CPX_OFF);
  if ( status != 0 ) {
    fprintf (stderr,"Failure to deactivate agregator indicator, error %d.\n", status);
    goto TERMINATE;
  }
  */

  /* deactivate the coefficient reduction indicator */  
  /*
  status = CPXsetintparam (env, CPX_PARAM_COEREDIND, CPX_OFF);
  if ( status != 0 ) {
    fprintf (stderr,"Failure to deactivate coefficient reduction indicator, error %d.\n", status);
    goto TERMINATE;
  }
  */

  /* Optimize the problem and obtain solution. */

  switch (argv[1][0]) {
  case 'd':
    status = CPXhybnetopt(env,lp,'t');
    break;
  case 'p':
    status = CPXhybnetopt(env,lp,'o');
    break;
  case 'b':
    status = CPXbaropt(env,lp);
    break;
  case 'o':
    status = CPXoptimize(env,lp);
    break;
  case 't':
    status = CPXdualopt(env,lp);
    break;
  default:
    status= -1;
    break;
  }
  if ( status ) {
    fprintf (stderr, "Failed to optimize LP.\n");
    goto TERMINATE;
   }

  solstat = CPXgetstat (env, lp);
  nit = CPXgetitc(env, lp);
  status  = CPXgetobjval (env, lp, &objval);
  if ( status ) {
    fprintf (stderr,"Failed to obtain objective value. solstat= %d\n",solstat);
    goto TERMINATE;
  }
  
  times(&buffer);
  cputime= (double)(buffer.tms_utime/(double)CLOCKS_PER_SECOND);
  if (argv[1][0] == 'b') fout= fopen("network_b.stat","a+");
  if (argv[1][0] == 'd') fout= fopen("network_d.stat","a+");  
  if (argv[1][0] == 'p') fout= fopen("network_p.stat","a+");
  if (argv[1][0] == 'o') fout= fopen("network_o.stat","a+");
  if (argv[1][0] == 't') fout= fopen("network_t.stat","a+");
  fprintf(fout,"Model= %-15s   Exit= %-4d   Obj.F.= %+1.11e   N.it.= %-4d   Cpu.sec.= %-11.2f  Solver= %c\n",argv[2],solstat,objval,nit,cputime,argv[1][0] );

  fclose(fout);

 TERMINATE:

  /* Free up the basis and solution */

  free_and_null ((char **) &cstat);
  free_and_null ((char **) &rstat);
  free_and_null ((char **) &x);

  /* Free up the problem as allocated by CPXloadlp, if necessary */

  if ( lp != NULL ) {
    status = CPXunloadprob (env, &lp);
    if ( status ) {
      fprintf (stderr, "CPXunloadprob failed, error code %d.\n", status);
    }
  }

  /* Free up the data read in by the readers */

/*  freereaddata (&probname, &obj, &rhs, &sense, &matbeg, 
		&matcnt, &matind, &matval, &lb, &ub,
		&rngval, &colname, &colnamestore, 
		&rowname, &rownamestore);
  */
  /* Free up the CPLEX environment, if necessary */
  
  if ( env != NULL ) {
    status = CPXcloseCPLEX (&env);
    
    /* Note that CPXcloseCPLEX produces no output,
       so the only way to see the cause of the error is to use
       CPXgeterrorstring.  For other CPLEX routines, the errors will
       be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON. */
    
    if ( status ) {
      char  errmsg[1024];
      fprintf (stderr, "Could not close CPLEX environment.\n");
      CPXgeterrorstring (env, status, errmsg);
      fprintf (stderr, "%s", errmsg);
    }
  }
  
  return (status);
  
} /* END main */


static void read_net_create_lp(int *numcols_p, int * numrows_p, int * objsen_p, 
			       double **obj_p, double **rhs_p, char **sense_p, 
			       int **matbeg_p, int **matcnt_p, int **matind_p, 
			       double **matval_p, double **lb_p, double **ub_p, 
			       double **rngval_p, int *colspace_p, int *rowspace_p, 
			       int *nzspace_p)
{
  /* Local variables for reading network problem */
  int nar,nat,nnu,ncb,nlb,*mnk= NULL,*mnl= NULL,*arccb= NULL,*artcb= NULL,*numcb= NULL;
  double *kap= NULL,*cost= NULL,*cmt= NULL,*inj= NULL,*lowcb= NULL,*upcb= NULL,*coefcb= NULL;
  double inf= 1.0e+32,raux;
  int i,j,k,nrhs,l,nv,ctlb,fivcb,ilb,nranged,iaux,ncmt,icmt;
  
  /* read network dimensions from fortran routine */
  read_dimen_pprn_(&nnu,&nar,&nat,&ncb,&nlb);

  /* allocate memory */
  if (ncb == 0) nlb=0;
  ALLOC(mnk,nar,int);
  ALLOC(mnl,nar,int);
  ALLOC(kap,nar*nat,double);
  ALLOC(cost,nar*nat,double);
  ALLOC(cmt,nar,double);
  ALLOC(inj,nnu*nat,double);
  if (ncb>0) {
    ALLOC(lowcb,ncb,double);
    ALLOC(upcb,ncb,double);
    ALLOC(arccb,nlb,int);
    ALLOC(artcb,nlb,int);
    ALLOC(numcb,nlb,int);
    ALLOC(coefcb,nlb,double);
  }
  
  /* read network from fortran routine and resize some arrays if required */
  read_format_pprn_(&nnu,&nar,&nat,&ncb,&nlb,cost,kap,cmt,inj,lowcb,upcb,
		    coefcb,arccb,artcb,numcb,mnk,mnl);
  if (ncb>0) {
    REALLOC(arccb,nlb,int);
    REALLOC(artcb,nlb,int);
    REALLOC(numcb,nlb,int);
    REALLOC(coefcb,nlb,double);
  }

  /* data read in fortran numeration are transformed to C numeration */
  for(i=0;i<nar;i++) {
    mnk[i]--;
    mnl[i]--;
  }
  for(i=0;i<nlb;i++) {
    arccb[i]--;
    artcb[i]--;
    numcb[i]--;
  }

  /* create linear problem structure; fill in linear problem arrays */
  
  /* dimensions and objsen */
  *objsen_p= 1;
  *numcols_p= nar*nat;
  if (nat == 1)  /* don't consider mutual capacity constraints */
    *numrows_p= nnu*nat+ncb;
  else  {/* consider mutual capacity constraints */
    for(i=0,ncmt=0;i<nar;i++) {
      if (cmt[i] < inf) ncmt++;
    }
    *numrows_p= nnu*nat+ncmt+ncb;
  }
  /* count ranged side constraints and initialize dimensions of some lp vectors */
  nranged= 0;
  for(i=0;i<ncb;i++) {
    if ((lowcb[i] != -inf)&&(upcb[i] != inf)&&(lowcb[i] != upcb[i])) nranged++;
  }
  *colspace_p= *numcols_p+nranged;
  *rowspace_p= *numrows_p;
  if (nat ==1)  /* don't consider mutual capacity constraints */
    *nzspace_p=  2*nar*nat+nlb+nranged;
  else  /* consider mutual capacity constraints */
    *nzspace_p=  2*nar*nat+ncmt*nat+nlb+nranged;
  
  /* obj */
  REALLOC(cost,(*colspace_p),double);
  *obj_p= cost;
  
  /* lb, lu */
  REALLOC(kap,(*colspace_p),double);
  ALLOC((*lb_p),(*colspace_p),double);
  *ub_p= kap;
  
  /* rhs, sense, rngval */
  REALLOC(inj,(*rowspace_p),double);
  *rhs_p= inj; /* first nnu*nat rhs positions have already the network demands/supplies */ 
  ALLOC((*sense_p),(*rowspace_p),char);    
  if (nranged == 0) 
    *rngval_p= NULL;
  else
    ALLOC((*rngval_p),(*rowspace_p),double);
  /* network constraints */
  for(i=0;i<nnu*nat;i++) (*sense_p)[i]= 'E';
  nrhs= nnu*nat;
  /* mutual capactity constraints */
  if (nat>1) {
    icmt= 0;
    for(i=0;i<nar;i++) {
      if (cmt[i] < inf) {
	(*rhs_p)[nrhs+icmt]= cmt[i];
	(*sense_p)[nrhs+icmt]= 'L';
	icmt++;
      }
    }
    nrhs= nrhs+ncmt;
  }
  /* side constraints */
  for(i=0;i<ncb;i++,nrhs++) {
    if (lowcb[i] == -inf ) {
      if (upcb[i] == inf) {
	fprintf(stderr, "warning: inactive side constraint %d\n",i);
	(*sense_p)[nrhs]= 'L';
	(*rhs_p)[nrhs]= inf;
      }
      else { /* upcb[i] != inf */
	(*sense_p)[nrhs]= 'L';
	(*rhs_p)[nrhs]= upcb[i];
      }
    }
    else { /* lowcb[i] != -inf */
      if (upcb[i] == inf) {
	(*sense_p)[nrhs]= 'G';
	(*rhs_p)[nrhs]= lowcb[i];
      }
      else { /* upcb[i] != inf */
	if (lowcb[i] == upcb[i]) {
	  (*sense_p)[nrhs]= 'E';
	  (*rhs_p)[nrhs]= lowcb[i];
	}
	else {
	  (*sense_p)[nrhs]= 'R';
	  (*rhs_p)[nrhs]= lowcb[i];
	  (*rngval_p)[nrhs]= upcb[i]-lowcb[i];
	}
      }
    }
  }
  
  /* matbeg, matcnt, matind, matval */
  ALLOC((*matbeg_p),(*colspace_p),int);
  ALLOC((*matcnt_p),(*colspace_p),int);
  ALLOC((*matind_p),(*nzspace_p),int);
  ALLOC((*matval_p),(*nzspace_p),double);
  /* count side constraints nonzero elements per variable in matrix constraints */
  for(i=0;i<nlb;i++) (*matcnt_p)[(nar*artcb[i])+arccb[i]]++;
  /* create matind and matval */
  nv= 0;
  ctlb= 0;
  l= 0;
  for (k=0;k<nat;k++){
    icmt= 0;
    for (j=0;j<nar;j++,nv++) {
      /* network constraints */
      (*matval_p)[l]= 1.0;  (*matind_p)[l]= k*nnu+mnk[j]; 
      l++;
      (*matval_p)[l]= -1.0; (*matind_p)[l]= k*nnu+mnl[j];
      l++;
      /* mutual capacity constraints */
      if ((nat>1)&&(cmt[j]<inf)) {
	(*matval_p)[l]= 1.0; (*matind_p)[l]= nnu*nat+icmt;
	l++;
	icmt++;
      }
      /* side constraints */
      fivcb= ctlb+(*matcnt_p)[nv];
      for(ilb=ctlb;(ilb<nlb)&&(ctlb<fivcb);ilb++) {
	if ((arccb[ilb]==j)&&(artcb[ilb]==k)) {
	  (*matval_p)[l]= coefcb[ilb]; 
	  if (nat>1) 
	    (*matind_p)[l]= nnu*nat+ncmt+numcb[ilb];
	  else
	    (*matind_p)[l]= nnu*nat+numcb[ilb];
	  l++;
	  SWAP(arccb[ilb],arccb[ctlb],iaux);
	  SWAP(artcb[ilb],artcb[ctlb],iaux);
	  SWAP(numcb[ilb],numcb[ctlb],iaux);
	  SWAP(coefcb[ilb],coefcb[ctlb],raux);
	  ctlb++;
	}
      }
    }
  }
  /* create matbeg and matcnt */
  i= 0;
  (*matbeg_p)[0]= 0;
  for (k=0;k<nat;k++)
    for (j=0;j<nar;j++,nv++) {
      (*matcnt_p)[i] += 2;
      if ((nat>1)&&(cmt[j]<inf)) (*matcnt_p)[i]++;
      (*matbeg_p)[i+1]= (*matbeg_p)[i]+(*matcnt_p)[i];
      i++;
    }
  
  /* free some network vectors --those not used  as lp vectors */
  free_and_null2((char *) mnk);
  free_and_null2((char *) mnl);
  free_and_null2((char *) cmt);
  free_and_null2((char *) lowcb);  
  free_and_null2((char *) upcb);
  free_and_null2((char *) arccb);
  free_and_null2((char *) artcb);
  free_and_null2((char *) numcb);
  free_and_null2((char *) coefcb);
} /* END read_net_create_lp */  

static void freereaddata (char **probname_p, double **obj_p, 
			  double **rhs_p, char **sense_p,
			  int **matbeg_p, int **matcnt_p, 
			  int **matind_p, double **matval_p, 
			  double **lb_p, double **ub_p, double **rngval_p,
			  char ***colname_p, char **colnamestore_p, 
			  char ***rowname_p, char **rownamestore_p)
{
  free_and_null ((char **) probname_p);
  free_and_null ((char **) obj_p);
  free_and_null ((char **) rhs_p);
  free_and_null ((char **) sense_p);
  free_and_null ((char **) matbeg_p);
  free_and_null ((char **) matcnt_p);
  free_and_null ((char **) matind_p);
  free_and_null ((char **) matval_p);
  free_and_null ((char **) lb_p);
  free_and_null ((char **) ub_p);
  free_and_null ((char **) rngval_p);
  free_and_null ((char **) colname_p);
  free_and_null ((char **) colnamestore_p);
  free_and_null ((char **) rowname_p);
  free_and_null ((char **) rownamestore_p);
} /* END freereaddata */


/* This simple routine frees up the pointer *ptr, and sets *ptr to NULL */

static void free_and_null (char **ptr)
{
  if ( *ptr != NULL ) {
    free (*ptr);
    *ptr = NULL;
  }
} /* END free_and_null */  

/* This simple routine frees up the pointer ptr, and sets *ptr to NULL */

static void free_and_null2 (char *ptr)
{
  if ( ptr != NULL ) {
    free (ptr);
    ptr = NULL;
  }
} /* END free_and_null2 */  

/* auxiliary routines for writing vectors */
static void  write_double(double *v, int i, int j, char *nom)
{
  int k;
  
  for(k=i; k<=j; k++) printf("%s(%i)= %e\n",nom,k,v[k]); 
}

static void write_int(int *v, int i, int j, char *nom)
{
  int k;
  
  for(k=i; k<=j; k++) printf("%s(%i)= %i\n",nom,k,v[k]); 
}
static void write_char(char *v, int i, int j, char *nom)
{
  int k;
  
  for(k=i; k<=j; k++) printf("%s(%i)= %c\n",nom,k,v[k]); 
}







